"""SWF decisions making."""

import abc
import typing as t

T = t.TypeVar("T")


class DecisionsBuilder(metaclass=abc.ABCMeta):
    """SWF decision builder.

    Args:
        workflow: workflow specification
        task: decision task
    """

    def __init__(self, workflow: "Workflow", task: t.Dict[str, t.Any]):
        self.workflow = workflow
        self.task = task
        self.decisions = []

    @abc.abstractmethod
    def build_decisions(self):  # pragma: no cover
        """Build decisions from workflow history."""
        raise NotImplementedError


class Workflow(metaclass=abc.ABCMeta):
    """SWF workflow specification.

    Args:
        name: workflow name
        version: workflow version
    """

    def __init__(self, name: str, version: str, description: str = None):
        self.name = name
        self.version = version
        self.description = description

    @classmethod
    def from_spec(cls: T, spec: t.Dict[str, t.Any]) -> T:
        """Construct workflow type from specification.

        Args:
            spec: workflow specification
        """

        return cls(spec["name"], spec["version"], spec.get("description"))

    @property
    @abc.abstractmethod
    def decisions_builder(self) -> DecisionsBuilder:  # pragma: no cover
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def spec_type(self) -> str:  # pragma: no cover
        raise NotImplementedError

    def setup(self):
        """Set up workflow specification.

        Useful for pre-calculation or other initialisation.
        """

    def make_decisions(self, task: t.Dict[str, t.Any]) -> t.List[t.Dict[str, t.Any]]:
        """Build decisions from workflow history.

        Args:
            task: decision task

        Returns:
            workflow decisions
        """

        builder = self.decisions_builder(self, task)
        builder.build_decisions()
        return builder.decisions



"""SWF decider."""

import json
import uuid
import socket
import pathlib
import typing as t
import logging as lg

import boto3

from . import _util
from . import decisions as seddy_decisions

logger = lg.getLogger(__name__)
socket.setdefaulttimeout(70.0)


class Decider:
    """SWF decider.

    Args:
        workflows: decider workflows
        domain: SWF domain to poll in
        task_list: SWF decider task-list

    Attributes:
        client (botocore.client.BaseClient): SWF client
        identity: name of decider to poll as
    """

    def __init__(
        self, workflows: t.List[seddy_decisions.Workflow], domain: str, task_list: str,
    ):
        self.workflows = workflows
        self.domain = domain
        self.task_list = task_list
        self.client = boto3.client("swf")
        self.identity = socket.getfqdn() + "-" + str(uuid.uuid4())[:8]

    def _poll_for_decision_task(self) -> t.Dict[str, t.Any]:
        """Poll for a decision task from SWF.

        See https://docs.aws.amazon.com/amazonswf/latest/apireference/API_PollForDecisionTask.html

        Returns:
            decision task
        """

        return _util.list_paginated(
            self.client.poll_for_decision_task,
            "events",
            {
                "domain": self.domain,
                "identity": self.identity,
                "taskList": {"name": self.task_list},
            },
        )

    def _make_decisions(self, task: t.Dict[str, t.Any]) -> t.List[t.Dict[str, t.Any]]:
        """Build decisions from workflow history.

        Args:
            task: decision task

        Returns:
            workflow decisions
        """

        workflow_ids = [(w.name, w.version) for w in self.workflows]
        task_id = (task["workflowType"]["name"], task["workflowType"]["version"])
        idx = workflow_ids.index(task_id)
        workflow = self.workflows[idx]
        return workflow.make_decisions(task)

    def _respond_decision_task_completed(
        self, decisions: t.List[t.Dict[str, t.Any]], task: t.Dict[str, t.Any]
    ):
        """Send decisions to SWF.

        See https://docs.aws.amazon.com/amazonswf/latest/apireference/API_RespondDecisionTaskCompleted.html

        Args:
            decisions: workflow decisions
            task: decision task
        """

        logger.debug(
            "Sending %d decisions for task '%s'", len(decisions), task["taskToken"]
        )
        self.client.respond_decision_task_completed(
            taskToken=task["taskToken"], decisions=decisions
        )

    def _poll_and_run(self):
        """Perform poll, and possibly run decision task."""
        task = self._poll_for_decision_task()
        if not task["taskToken"]:
            return
        logger.log(
            25,
            "Got decision task '%s' for workflow '%s-%s' execution '%s' (run '%s')",
            task["taskToken"],
            task["workflowType"]["name"],
            task["workflowType"]["version"],
            task["workflowExecution"]["workflowId"],
            task["workflowExecution"]["runId"],
        )
        decisions = self._make_decisions(task)
        self._respond_decision_task_completed(decisions, task)

    def _run_uncaught(self):
        """Run decider."""
        logger.log(
            25,
            "Polling for tasks in domain '%s' with task-list '%s' as '%s'",
            self.domain,
            self.task_list,
            self.identity,
        )
        while True:
            self._poll_and_run()

    def run(self):
        """Run decider."""
        try:
            self._run_uncaught()
        except KeyboardInterrupt:
            logger.info("Quitting due to keyboard-interrupt")


def run_app(workflows_spec_json: pathlib.Path, domain: str, task_list: str):
    """Run decider application.

    Arguments:
        workflows_spec_json: workflows specifications JSON
        domain: SWF domain
        task_list: SWF decider task-list
    """

    decider_spec = json.loads(workflows_spec_json.read_text())
    workflows = _util.construct_workflows(decider_spec)
    _util.setup_workflows(workflows)
    decider = Decider(workflows, domain, task_list)
    decider.run()



"""SWF decider service utilities."""

import typing as t
import logging as lg

from . import decisions as seddy_decisions

LOGGING_LEVELS = {
    -2: lg.ERROR,
    -1: lg.WARNING,
    0: 25,
    1: lg.INFO,
    2: lg.DEBUG,
}


def setup_logging(verbose: int):
    """Setup logging.

    Args:
        verbose: logging verbosity
    """

    lg.addLevelName(25, "NOTICE")
    level = LOGGING_LEVELS.get(verbose, lg.CRITICAL if verbose < 0 else lg.NOTSET)
    fmt = "%(asctime)s [%(levelname)8s] %(name)s: %(message)s"

    try:
        import coloredlogs
    except ImportError:
        lg.basicConfig(level=level, format=fmt)
        return

    coloredlogs.install(
        level=level,
        fmt=fmt,
        field_styles={
            "asctime": {"faint": True, "color": "white"},
            "levelname": {"bold": True, "color": "blue"},
            "name": {"bold": True, "color": "yellow"},
        },
        level_styles={
            **coloredlogs.DEFAULT_LEVEL_STYLES,
            "notice": {},
            "info": {"color": "white"},
        },
        milliseconds=True,
    )
    lg.root.setLevel(level)


def list_paginated(
    fn: t.Callable[..., t.Dict[str, t.Any]],
    list_key: str,
    kwargs: t.Dict[str, t.Any] = None,
    next_key: str = "nextPageToken",
    next_arg: str = None,
) -> t.Dict[str, t.Any]:
    """List AWS resources, consuming pagination.

    Args:
        fn: resource listing function
        list_key: key of paginated list in response
        kwargs: keyword arguments to ``fn``
        next_key: key of next-page token in response
        next_arg: argument name of next-page token in ``fn``, default: same
            as ``next_key``

    Returns:
        collected response of ``fn``
    """

    next_arg = next_key if next_arg is None else next_arg
    kwargs = kwargs or {}
    resp = fn(**kwargs)
    if resp.get(next_key):
        kwargs = kwargs.copy()
        kwargs[next_arg] = resp.pop(next_key)
        new_resp = list_paginated(fn, list_key, kwargs, next_key, next_arg)
        resp[list_key].extend(new_resp[list_key])
    return resp


def construct_workflows(
    workflows_spec: t.Dict[str, t.Any]
) -> t.List[seddy_decisions.Workflow]:
    """Construct workflows from specification.

    Args:
        workflows_spec: workflows specifications

    Returns:
        workflow type specifications
    """

    assert (1,) < tuple(map(int, workflows_spec["version"].split("."))) < (2,)
    workflows = []
    for workflow_spec in workflows_spec["workflows"]:
        workflow_cls = seddy_decisions.WORKFLOW[workflow_spec["spec_type"]]
        workflow = workflow_cls.from_spec(workflow_spec)
        workflows.append(workflow)
    return workflows


def setup_workflows(workflows: t.List[seddy_decisions.Workflow]):
    """Set-up decider workflows.

    Args:
        workflows: workflow type specifications

    Returns:
        decider initialised workflows
    """

    for workflow in workflows:
        workflow.setup()



"""SWF decisions building."""

__all__ = ["DecisionsBuilder", "Workflow", "DAGBuilder", "DAGWorkflow"]

from ._base import DecisionsBuilder
from ._base import Workflow
from ._dag import DAGBuilder
from ._dag import DAGWorkflow

WORKFLOW = {
    DAGWorkflow.spec_type: DAGWorkflow,
}



"""SWF decisions making."""

import json
import typing as t

from . import _base

_at_attr_keys = {
    "ActivityTaskCompleted": "activityTaskCompletedEventAttributes",
    "ActivityTaskFailed": "activityTaskFailedEventAttributes",
    "ActivityTaskTimedOut": "activityTaskTimedOutEventAttributes",
    "ActivityTaskScheduled": "activityTaskScheduledEventAttributes",
    "ActivityTaskStarted": "activityTaskStartedEventAttributes",
}


def _get(item_id, items, id_key):
    """Get item from list with given ID."""
    return next(item for item in items if item[id_key] == item_id)


class DAGBuilder(_base.DecisionsBuilder):
    """SWF decision builder from DAG-type workflow specification."""

    def __init__(self, workflow: "DAGWorkflow", task):
        super().__init__(workflow, task)
        self._scheduled = {}
        self._activity_task_events = {at["id"]: [] for at in workflow.task_specs}

    def _schedule_task(
        self,
        activity_task: t.Dict[str, t.Any],
        workflow_started_event: t.Dict[str, t.Any],
    ):
        assert workflow_started_event["eventType"] == "WorkflowExecutionStarted"
        attrs = workflow_started_event["workflowExecutionStartedEventAttributes"]
        decision_attributes = {
            "activityId": activity_task["id"],
            "activityType": activity_task["type"],
        }
        input_ = json.loads(attrs.get("input", "null"))
        if input_:
            decision_attributes["input"] = json.dumps(input_[activity_task["id"]])
        if "heartbeat" in activity_task:
            decision_attributes["heartbeatTimeout"] = str(activity_task["heartbeat"])
        if "timeout" in activity_task:
            decision_attributes["startToCloseTimeout"] = str(activity_task["timeout"])
        if "task_list" in activity_task:
            decision_attributes["taskList"] = activity_task["task_list"]
        if "priority" in activity_task:
            decision_attributes["taskPriority"] = str(activity_task["priority"])
        self.decisions.append(
            {
                "decisionType": "ScheduleActivityTask",
                "scheduleActivityTaskDecisionAttributes": decision_attributes,
            }
        )

    def _get_scheduled_references(self):
        for event in self.task["events"]:
            if event["eventType"] in _at_attr_keys:
                if event["eventType"] == "ActivityTaskScheduled":
                    self._scheduled[event["eventId"]] = event
                else:
                    self._scheduled[event["eventId"]] = _get(
                        event[_at_attr_keys[event["eventType"]]]["scheduledEventId"],
                        self.task["events"],
                        "eventId",
                    )

    def _get_activity_task_events(self):
        for event in self.task["events"]:
            if event["eventType"] in _at_attr_keys:
                scheduled_event = self._scheduled[event["eventId"]]
                attrs = scheduled_event["activityTaskScheduledEventAttributes"]
                self._activity_task_events[attrs["activityId"]].append(event)

    def _process_activity_task_completed_event(self, event: t.Dict[str, t.Any]):
        # Schedule dependants
        scheduled_event = self._scheduled[event["eventId"]]
        attrs = scheduled_event["activityTaskScheduledEventAttributes"]
        dependants_task = self.workflow.dependants[attrs["activityId"]]
        for activity_task_id in dependants_task:
            assert not self._activity_task_events[activity_task_id]
            activity_task = _get(activity_task_id, self.workflow.task_specs, "id")
            dependencies_satisfied = True
            for dependency_activity_task_id in activity_task["dependencies"]:
                events = self._activity_task_events[dependency_activity_task_id]
                if not events or events[-1]["eventType"] != "ActivityTaskCompleted":
                    dependencies_satisfied = False
                    break
            if dependencies_satisfied:
                self._schedule_task(activity_task, self.task["events"][0])

        # Complete workflow
        tasks_complete = True
        for events in self._activity_task_events.values():
            if not events or events[-1]["eventType"] != "ActivityTaskCompleted":
                tasks_complete = False
                break
        if tasks_complete:
            result = {}
            for activity_id, events in self._activity_task_events.items():
                assert events and events[-1]["eventType"] == "ActivityTaskCompleted"
                attrs = events[-1].get("activityTaskCompletedEventAttributes")
                if attrs and "result" in attrs:
                    result[activity_id] = json.loads(attrs["result"])
            decision = {"decisionType": "CompleteWorkflowExecution"}
            if result:
                decision["completeWorkflowExecutionDecisionAttributes"] = {
                    "result": json.dumps(result)
                }
            self.decisions = [decision]
            return True

    def _process_activity_task_failed_event(self, event: t.Dict[str, t.Any]):
        attr = self._scheduled[event["eventId"]]["activityTaskScheduledEventAttributes"]
        self.decisions = [
            {
                "decisionType": "FailWorkflowExecution",
                "failWorkflowExecutionDecisionAttributes": {
                    "reason": "activityFailure",
                    "details": "Activity '%s' failed" % attr["activityId"],
                },
            },
        ]

    def _process_activity_task_timed_out_event(self, event: t.Dict[str, t.Any]):
        attr = self._scheduled[event["eventId"]]["activityTaskScheduledEventAttributes"]
        self.decisions = [
            {
                "decisionType": "FailWorkflowExecution",
                "failWorkflowExecutionDecisionAttributes": {
                    "reason": "activityTimeOut",
                    "details": "Activity '%s' timed-out" % attr["activityId"],
                },
            },
        ]

    def _process_cancel_requested_event(self):
        decisions = []
        for activity_task in self.workflow.task_specs:
            events = self._activity_task_events[activity_task["id"]]
            if events and events[-1]["eventType"] in (
                "ActivityTaskScheduled",
                "ActivityTaskStarted",
            ):
                decisions.append(
                    {
                        "decisionType": "RequestCancelActivityTask",
                        "requestCancelActivityTaskDecisionAttributes": {
                            "activityId": activity_task["id"]
                        },
                    }
                )
        decisions.append({"decisionType": "CancelWorkflowExecution"})
        self.decisions = decisions

    def _process_workflow_execution_started_event(self, event: t.Dict[str, t.Any]):
        assert event is self.task["events"][0]
        for activity_task in self.workflow.task_specs:
            assert not self._activity_task_events[activity_task["id"]]
            if not activity_task.get("dependencies"):
                self._schedule_task(activity_task, event)

    def _process_new_events(self):
        event_ids = [event["eventId"] for event in self.task["events"]]
        current_idx = event_ids.index(self.task["startedEventId"])
        previous_idx = -1
        if self.task["previousStartedEventId"] in event_ids:
            previous_idx = event_ids.index(self.task["previousStartedEventId"])
        events = self.task["events"][previous_idx + 1:current_idx + 1]
        assert self.task["events"][-1]["eventType"] == "DecisionTaskStarted"
        assert self.task["events"][-2]["eventType"] == "DecisionTaskScheduled"
        for event in events[:-2]:
            if event["eventType"] == "ActivityTaskCompleted":
                if self._process_activity_task_completed_event(event):
                    break
            elif event["eventType"] == "ActivityTaskFailed":
                self._process_activity_task_failed_event(event)
                break
            elif event["eventType"] == "ActivityTaskTimedOut":
                self._process_activity_task_timed_out_event(event)
                break
            elif event["eventType"] == "WorkflowExecutionCancelRequested":
                self._process_cancel_requested_event()
                break
            elif event["eventType"] == "WorkflowExecutionStarted":
                self._process_workflow_execution_started_event(event)

    def build_decisions(self):
        self._get_scheduled_references()
        self._get_activity_task_events()
        self._process_new_events()


class DAGWorkflow(_base.Workflow):
    """Dag-type SWF workflow specification.

    Args:
        name: workflow name
        version: workflow version
        task_specs: DAG task specifications
    """

    spec_type = "dag"
    decisions_builder = DAGBuilder

    def __init__(
        self, name, version, task_specs: t.List[t.Dict[str, t.Any]], description=None
    ):
        super().__init__(name, version, description)
        self.task_specs = task_specs
        self.dependants = {}

    @classmethod
    def from_spec(cls, spec):
        return cls(
            spec["name"], spec["version"], spec["tasks"], spec.get("description")
        )

    def _build_dependants(self):
        for activity_task in self.task_specs:
            dependants_task = []
            for other_activity_task in self.task_specs:
                if activity_task["id"] in other_activity_task.get("dependencies", []):
                    dependants_task.append(other_activity_task["id"])
            self.dependants[activity_task["id"]] = dependants_task

    def setup(self):
        self._build_dependants()
